#ifndef TSYNC_OBJECT_H
#define TSYNC_OBJECT_H

/////////////////////////////////////////////////////////////////////////////
//
// File: TsynObject.h
// Author: Kelvin Sung
// Date: April 2003
//
// 		TsyncObject: A class demonstrating some of the functionality of
// 		the MFC synchronization objects (CEvent/CMutex).
//
// **Note** = 
//
//			Mutex Locks are per-Thread property, once a thread obtains a lock
//			only that thread can "unlock", and as long as a thread has the lock
//			to a mutex, no other threads should try to unlock (this is a run time error)
//			
//
//			Event Locks are waiting for signals, (SetEvent, or PulseEvent).
//			No one thread really owns the Lock, all threads waiting on an event
//			Lock has the potential being wake up when any thread Set/PulseEvent
//			Whether all or one waiting thread gets wake up by an event depends
//			on if the CEvent is mutual or automatic reset (refer to MSDN on-line
//			help pages for more about this).
//
//	When Event is present (Think Green light) - do not block on Lock (think do not stop at intersection)
//	When NoEvent (Re-set) (Think Red Light)   - block on Lock (think stop at intersection)
//
//
// ***CMutex***
// ============
//		Changing state/inquiry:
//
//			Lock() - if its not already locked, 
//							you "get" the lock and no-one else can get it
//							until you "unlock"
//				     if its already locked,
//							you get blocked, until the thread the locks does a "unlock"
//
//			unlock() - you can only unlock a thread you have locked.
//					   this will wake up _ALL_ waiting threads (those who tried to lock)
//					   and only one thread will get the next lock
//			
//
// ***CEvent***
// ============
//
//		Changing state:
//
//			SetEvent() - Event present (Green Light), all blocked threads can go now
//			ResetEvent() - Remove Event (Red Light), all future Lock() gets blocked
//
//
//		Inquiring about state:
//
//			Lock() - check for the presence of event (at intersection)
//					 if event is present (someone did a SetEvent()), thread can continue
//					 if event is not present, this will block the thread
//


#include <iostream.h>

#include <afxmt.h>
		// for MFC Synchronization objects (CEvent/CMutex etc.)

#define kAllThreadQuits -10

class TsyncObject {
	private:
		int			fCommand;	
						// for now, quit signal is kept here ...

		int			fTotalWorkTime;
						// total time of work done by all workers
		CMutex		*fWorkTimeLock;
						// make sure only one thread changes the fTotalWorkTime at anytime

		CEvent  	*fStudentWorkEvent;
						// Kelvin Thread will reset this event when it wants to work
						// All student threads should check this the event (by trying to lock it),
						// if the event is set, then student threads can work, otherwise,
						// student threads should wait for this even to become set.
		
	public:
		
		TsyncObject() : fCommand(0), fTotalWorkTime(0)
			{
				fWorkTimeLock = new CMutex();
				fWorkTimeLock->Unlock();
					// initially no one is changing the fTotalWorkTime

				//
				// fCommand Mutex/Lock set ...
				fStudentWorkEvent = new CEvent(true,	// initially event is set
											true); 		// event must be reset manually. (instead of being a pulse)

				// initially StudentWorkEvent is set
				// this means initially student threads can work
			}

		
		void UpdateWorkTime(int t) {
				fWorkTimeLock->Lock();
				{
					int cpuRegister = fTotalWorkTime;
					Sleep(100);
					cpuRegister += t;
					fTotalWorkTime = cpuRegister;
					//
					// The above is simulating a very slow CPU implementing
					//
					//		 fTotalWorkTime += t;
					//
					//  where, fTotalWorkTime is first load into the CPU, then t is
					//  loaded in, then add, then results is stored back into fTotalWorkTime
					//  (in the memory).
				}
				fWorkTimeLock->Unlock();

				// Notice the function of fWorkTimeLock is to "protect" the access to
				// fTotalWorkTime, we need this protection anytime 2 or more thread have
				// access to change a variable. When both threads attempts to increment
				// the variable at the same time, the outcome becomes un-preditable, it is
				// possible for the result to be wrong!!
		}
		int TotalWorkTime() { return fTotalWorkTime; }

		void KelvinStartWork() 	{
			fStudentWorkEvent->ResetEvent();
				// Kelvin Thread starts working now ... clear studentWorkEvent
				// anyone who tries to perform StudentWorkEvent->Lock() will get blocked
		}

		void KelvinEndWork() {
			fStudentWorkEvent->SetEvent();
				// Kelvin Thread done, 
				//   1. All currently locked threads (StudentWorkEvent->Lock()) will be wake up
				//   2. Any future StudentWorkEventEvent->Lock() will _NOT_ cause locking
		}
			
		void StudentDoWork()
			// Student thread comes to check if it can do work
			// if fKelvinEvent is set, this student thread will be
			// blocked until KelvinThread is done.
		{
			fStudentWorkEvent->Lock();
				// If there the event of fStudentWorkEvent is set, this function
				// will not do anything, if fStudentWorkEvent is cleared, then
				// caller thread gets block until fStudentWorkEvent is set.
		}

		void SetTerminateAll() { fCommand = kAllThreadQuits; }
		bool ShouldQuit() { return fCommand == kAllThreadQuits; }

};

#endif